@@ -13,23 +13,38 @@ used, there are several specialized subclasses of it.
     ~simpy.events.AllOf
 
 """
-from simpy._compat import PY2
-from simpy.exceptions import Interrupt, StopProcess
+from types import FrameType
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Generator,
+    Iterable,
+    Iterator,
+    List,
+    NewType,
+    Optional,
+    Tuple,
+)
 
-if PY2:
-    import sys
+from simpy.exceptions import Interrupt
 
+if TYPE_CHECKING and TYPE_CHECKING != 'SPHINX':  # Avoid circular import
+    from simpy.core import Environment, SimTime
 
-PENDING = object()
+PENDING: object = object()
 """Unique object to identify pending values of events."""
 
-URGENT = 0
+EventPriority = NewType('EventPriority', int)
+
+URGENT: EventPriority = EventPriority(0)
 """Priority of interrupts and process initialization events."""
-NORMAL = 1
+NORMAL: EventPriority = EventPriority(1)
 """Default priority used by events."""
 
 
-class Event(object):
+class Event:
     """An event that may happen at some point in time.
 
     An event
@@ -59,36 +74,40 @@ class Event(object):
     of them.
 
     """
-    def __init__(self, env):
+
+    _ok: bool
+    _defused: bool
+    _value: Any = PENDING
+
+    def __init__(self, env: 'Environment'):
         self.env = env
         """The :class:`~simpy.core.Environment` the event lives in."""
-        self.callbacks = []
+        self.callbacks: EventCallbacks = []
         """List of functions that are called when the event is processed."""
-        self._value = PENDING
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         """Return the description of the event (see :meth:`_desc`) with the id
         of the event."""
-        return '<%s object at 0x%x>' % (self._desc(), id(self))
+        return f'<{self._desc()} object at {id(self):#x}>'
 
-    def _desc(self):
+    def _desc(self) -> str:
         """Return a string *Event()*."""
-        return '%s()' % self.__class__.__name__
+        return f'{self.__class__.__name__}()'
 
     @property
-    def triggered(self):
+    def triggered(self) -> bool:
         """Becomes ``True`` if the event has been triggered and its callbacks
         are about to be invoked."""
         return self._value is not PENDING
 
     @property
-    def processed(self):
+    def processed(self) -> bool:
         """Becomes ``True`` if the event has been processed (e.g., its
         callbacks have been invoked)."""
         return self.callbacks is None
 
     @property
-    def ok(self):
+    def ok(self) -> bool:
         """Becomes ``True`` when the event has been triggered successfully.
 
         A "successful" event is one triggered with :meth:`succeed()`.
@@ -99,7 +118,7 @@ class Event(object):
         return self._ok
 
     @property
-    def defused(self):
+    def defused(self) -> bool:
         """Becomes ``True`` when the failed event's exception is "defused".
 
         When an event fails (i.e. with :meth:`fail()`), the failed event's
@@ -116,11 +135,11 @@ class Event(object):
         return hasattr(self, '_defused')
 
     @defused.setter
-    def defused(self, value):
+    def defused(self, value: bool) -> None:
         self._defused = True
 
     @property
-    def value(self):
+    def value(self) -> Optional[Any]:
         """The value of the event if it is available.
 
         The value is available when the event has been triggered.
@@ -129,10 +148,10 @@ class Event(object):
 
         """
         if self._value is PENDING:
-            raise AttributeError('Value of %s is not yet available' % self)
+            raise AttributeError(f'Value of {self} is not yet available')
         return self._value
 
-    def trigger(self, event):
+    def trigger(self, event: 'Event') -> None:
         """Trigger the event with the state and value of the provided *event*.
         Return *self* (this event instance).
 
@@ -144,7 +163,7 @@ class Event(object):
         self._value = event._value
         self.env.schedule(self)
 
-    def succeed(self, value=None):
+    def succeed(self, value: Optional[Any] = None) -> 'Event':
         """Set the event's value, mark it as successful and schedule it for
         processing by the environment. Returns the event instance.
 
@@ -152,14 +171,14 @@ class Event(object):
 
         """
         if self._value is not PENDING:
-            raise RuntimeError('%s has already been triggered' % self)
+            raise RuntimeError(f'{self} has already been triggered')
 
         self._ok = True
         self._value = value
         self.env.schedule(self)
         return self
 
-    def fail(self, exception):
+    def fail(self, exception: Exception) -> 'Event':
         """Set *exception* as the events value, mark it as failed and schedule
         it for processing by the environment. Returns the event instance.
 
@@ -169,50 +188,60 @@ class Event(object):
 
         """
         if self._value is not PENDING:
-            raise RuntimeError('%s has already been triggered' % self)
+            raise RuntimeError(f'{self} has already been triggered')
         if not isinstance(exception, BaseException):
-            raise ValueError('%s is not an exception.' % exception)
+            raise ValueError(f'{exception} is not an exception.')
         self._ok = False
         self._value = exception
         self.env.schedule(self)
         return self
 
-    def __and__(self, other):
+    def __and__(self, other: 'Event') -> 'Condition':
         """Return a :class:`~simpy.events.Condition` that will be triggered if
         both, this event and *other*, have been processed."""
         return Condition(self.env, Condition.all_events, [self, other])
 
-    def __or__(self, other):
+    def __or__(self, other: 'Event') -> 'Condition':
         """Return a :class:`~simpy.events.Condition` that will be triggered if
         either this event or *other* have been processed (or even both, if they
         happened concurrently)."""
         return Condition(self.env, Condition.any_events, [self, other])
 
 
+EventCallback = Callable[[Event], None]
+EventCallbacks = List[EventCallback]
+
+
 class Timeout(Event):
     """A :class:`~simpy.events.Event` that gets triggered after a *delay* has
     passed.
 
     This event is automatically triggered when it is created.
 
+
     """
-    def __init__(self, env, delay, value=None):
+
+    def __init__(
+        self,
+        env: 'Environment',
+        delay: 'SimTime',
+        value: Optional[Any] = None,
+    ):
         if delay < 0:
-            raise ValueError('Negative delay %s' % delay)
+            raise ValueError(f'Negative delay {delay}')
         # NOTE: The following initialization code is inlined from
         # Event.__init__() for performance reasons.
         self.env = env
-        self.callbacks = []
+        self.callbacks: EventCallbacks = []
         self._value = value
         self._delay = delay
         self._ok = True
         env.schedule(self, NORMAL, delay)
 
-    def _desc(self):
+    def _desc(self) -> str:
         """Return a string *Timeout(delay[, value=value])*."""
-        return '%s(%s%s)' % (self.__class__.__name__, self._delay,
-                             '' if self._value is None else
-                             (', value=%s' % self._value))
+        value_str = '' if self._value is None else f', value={self.value}'
+        return f'{self.__class__.__name__}({self._delay}{value_str})'
 
 
 class Initialize(Event):
@@ -221,12 +250,13 @@ class Initialize(Event):
     This event is automatically triggered when it is created.
 
     """
-    def __init__(self, env, process):
+
+    def __init__(self, env: 'Environment', process: 'Process'):
         # NOTE: The following initialization code is inlined from
         # Event.__init__() for performance reasons.
         self.env = env
-        self.callbacks = [process._resume]
-        self._value = None
+        self.callbacks: EventCallbacks = [process._resume]
+        self._value: Any = None
 
         # The initialization events needs to be scheduled as urgent so that it
         # will be handled before interrupts. Otherwise a process whose
@@ -242,18 +272,20 @@ class Interruption(Event):
     This event is automatically triggered when it is created.
 
     """
-    def __init__(self, process, cause):
+
+    def __init__(self, process: 'Process', cause: Optional[Any]):
         # NOTE: The following initialization code is inlined from
         # Event.__init__() for performance reasons.
         self.env = process.env
-        self.callbacks = [self._interrupt]
+        self.callbacks: EventCallbacks = [self._interrupt]
         self._value = Interrupt(cause)
         self._ok = False
         self._defused = True
 
         if process._value is not PENDING:
-            raise RuntimeError('%s has terminated and cannot be interrupted.' %
-                               process)
+            raise RuntimeError(
+                f'{process} has terminated and cannot be interrupted.'
+            )
 
         if process is self.env.active_process:
             raise RuntimeError('A process is not allowed to interrupt itself.')
@@ -261,7 +293,7 @@ class Interruption(Event):
         self.process = process
         self.env.schedule(self, URGENT)
 
-    def _interrupt(self, event):
+    def _interrupt(self, event: Event) -> None:
         # Ignore dead processes. Multiple concurrently scheduled interrupts
         # cause this situation. If the process dies while handling the first
         # one, the remaining interrupts must be ignored.
@@ -275,6 +307,9 @@ class Interruption(Event):
         self.process._resume(self)
 
 
+ProcessGenerator = Generator[Event, Any, Any]
+
+
 class Process(Event):
     """Process an event yielding generator.
 
@@ -287,45 +322,38 @@ class Process(Event):
     returns or raises an exception. The value of the process is the return
     value of the generator or the exception, respectively.
 
-    .. note::
-
-       Python version prior to 3.3 do not support return statements in
-       generators. You can use :meth:~simpy.core.Environment.exit() as
-       a workaround.
-
     Processes can be interrupted during their execution by :meth:`interrupt`.
 
     """
-    def __init__(self, env, generator):
+
+    def __init__(self, env: 'Environment', generator: ProcessGenerator):
         if not hasattr(generator, 'throw'):
             # Implementation note: Python implementations differ in the
             # generator types they provide. Cython adds its own generator type
             # in addition to the CPython type, which renders a type check
             # impractical. To workaround this issue, we check for attribute
             # name instead of type and optimistically assume that all objects
-            # with a ``throw`` attribute are generators (the more intuitive
-            # name ``__next__`` cannot be used because it was renamed from
-            # ``next`` in Python 2).
+            # with a ``throw`` attribute are generators.
             # Remove this workaround if it causes issues in production!
-            raise ValueError('%s is not a generator.' % generator)
+            raise ValueError(f'{generator} is not a generator.')
 
         # NOTE: The following initialization code is inlined from
         # Event.__init__() for performance reasons.
         self.env = env
-        self.callbacks = []
-        self._value = PENDING
+        self.callbacks: EventCallbacks = []
 
         self._generator = generator
 
         # Schedule the start of the execution of the process.
-        self._target = Initialize(env, self)
+        self._target: Event = Initialize(env, self)
 
-    def _desc(self):
+    def _desc(self) -> str:
         """Return a string *Process(process_func_name)*."""
-        return '%s(%s)' % (self.__class__.__name__, self._generator.__name__)
+        gen_name: str = self._generator.__name__  # type: ignore
+        return f'{self.__class__.__name__}({gen_name})'
 
     @property
-    def target(self):
+    def target(self) -> Event:
         """The event that the process is currently waiting for.
 
         Returns ``None`` if the process is dead or it is currently being
@@ -335,11 +363,11 @@ class Process(Event):
         return self._target
 
     @property
-    def is_alive(self):
+    def is_alive(self) -> bool:
         """``True`` until the process generator exits."""
         return self._value is PENDING
 
-    def interrupt(self, cause=None):
+    def interrupt(self, cause: Optional[Any] = None) -> None:
         """Interupt this process optionally providing a *cause*.
 
         A process cannot be interrupted if it already terminated. A process can
@@ -349,7 +377,7 @@ class Process(Event):
         """
         Interruption(self, cause)
 
-    def _resume(self, event):  # noqa: C901
+    def _resume(self, event: Event) -> None:  # noqa: C901
         """Resumes the execution of the process with the value of *event*. If
         the process generator exits, the process itself will get triggered with
         the return value or the exception of the generator."""
@@ -371,25 +399,21 @@ class Process(Event):
                     # processes.
                     exc = type(event._value)(*event._value.args)
                     exc.__cause__ = event._value
-                    if PY2:
-                        if hasattr(event._value, '__traceback__'):
-                            exc.__traceback__ = event._value.__traceback__
                     event = self._generator.throw(exc)
-            except (StopIteration, StopProcess) as e:
+            except StopIteration as e:
                 # Process has terminated.
-                event = None
+                event = None  # type: ignore
                 self._ok = True
                 self._value = e.args[0] if len(e.args) else None
                 self.env.schedule(self)
                 break
             except BaseException as e:
                 # Process has failed.
-                event = None
+                event = None  # type: ignore
                 self._ok = False
-                tb = e.__traceback__ if not PY2 else sys.exc_info()[2]
                 # Strip the frame of this function from the traceback as it
                 # does not add any useful information.
-                e.__traceback__ = tb.tb_next
+                e.__traceback__ = e.__traceback__.tb_next  # type: ignore
                 self._value = e
                 self.env.schedule(self)
                 break
@@ -405,11 +429,12 @@ class Process(Event):
             except AttributeError:
                 # Our optimism didn't work out, figure out what went wrong and
                 # inform the user.
-                if not hasattr(event, 'callbacks'):
-                    msg = 'Invalid yield value "%s"' % event
+                if hasattr(event, 'callbacks'):
+                    raise
 
+                msg = f'Invalid yield value "{event}"'
                 descr = _describe_frame(self._generator.gi_frame)
-                error = RuntimeError('\n%s%s' % (descr, msg))
+                error = RuntimeError(f'\n{descr}{msg}')
                 # Drop the AttributeError as the cause for this exception.
                 error.__cause__ = None
                 raise error
@@ -418,45 +443,47 @@ class Process(Event):
         self.env._active_proc = None
 
 
-class ConditionValue(object):
+class ConditionValue:
     """Result of a :class:`~simpy.events.Condition`. It supports convenient
     dict-like access to the triggered events and their values. The events are
     ordered by their occurences in the condition."""
 
     def __init__(self):
-        self.events = []
+        self.events: List[Event] = []
 
-    def __getitem__(self, key):
+    def __getitem__(self, key: Event) -> Any:
         if key not in self.events:
             raise KeyError(str(key))
 
         return key._value
 
-    def __contains__(self, key):
+    def __contains__(self, key: Event) -> bool:
         return key in self.events
 
-    def __eq__(self, other):
-        if type(other) is ConditionValue:
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, ConditionValue):
             return self.events == other.events
+        elif isinstance(other, dict):
+            return self.todict() == other
+        else:
+            return NotImplemented
 
-        return self.todict() == other
+    def __repr__(self) -> str:
+        return f'<ConditionValue {self.todict()}>'
 
-    def __repr__(self):
-        return '<ConditionValue %s>' % self.todict()
-
-    def __iter__(self):
+    def __iter__(self) -> Iterator[Event]:
         return self.keys()
 
-    def keys(self):
+    def keys(self) -> Iterator[Event]:
         return (event for event in self.events)
 
-    def values(self):
+    def values(self) -> Iterator[Any]:
         return (event._value for event in self.events)
 
-    def items(self):
+    def items(self) -> Iterator[Tuple[Event, Any]]:
         return ((event, event._value) for event in self.events)
 
-    def todict(self):
+    def todict(self) -> Dict[Event, Any]:
         return dict((event, event._value) for event in self.events)
 
 
@@ -481,10 +508,16 @@ class Condition(Event):
     Condition events can be nested.
 
     """
-    def __init__(self, env, evaluate, events):
-        super(Condition, self).__init__(env)
+
+    def __init__(
+        self,
+        env: 'Environment',
+        evaluate: Callable[[Tuple[Event, ...], int], bool],
+        events: Iterable[Event],
+    ):
+        super().__init__(env)
         self._evaluate = evaluate
-        self._events = events if type(events) is tuple else tuple(events)
+        self._events = tuple(events)
         self._count = 0
 
         if not self._events:
@@ -495,8 +528,10 @@ class Condition(Event):
         # Check if events belong to the same environment.
         for event in self._events:
             if self.env != event.env:
-                raise ValueError('It is not allowed to mix events from '
-                                 'different environments')
+                raise ValueError(
+                    'It is not allowed to mix events from different '
+                    'environments'
+                )
 
         # Check if the condition is met for each processed event. Attach
         # _check() as a callback otherwise.
@@ -508,14 +543,17 @@ class Condition(Event):
 
         # Register a callback which will build the value of this condition
         # after it has been triggered.
+        assert isinstance(self.callbacks, list)
         self.callbacks.append(self._build_value)
 
-    def _desc(self):
+    def _desc(self) -> str:
         """Return a string *Condition(evaluate, [events])*."""
-        return '%s(%s, %s)' % (self.__class__.__name__,
-                               self._evaluate.__name__, self._events)
+        return (
+            f'{self.__class__.__name__}('
+            f'{self._evaluate.__name__}, {self._events})'
+        )
 
-    def _populate_value(self, value):
+    def _populate_value(self, value: ConditionValue) -> None:
         """Populate the *value* by recursively visiting all nested
         conditions."""
 
@@ -525,14 +563,14 @@ class Condition(Event):
             elif event.callbacks is None:
                 value.events.append(event)
 
-    def _build_value(self, event):
+    def _build_value(self, event: Event) -> None:
         """Build the value of this condition."""
         self._remove_check_callbacks()
         if event._ok:
             self._value = ConditionValue()
             self._populate_value(self._value)
 
-    def _remove_check_callbacks(self):
+    def _remove_check_callbacks(self) -> None:
         """Remove _check() callbacks from events recursively.
 
         Once the condition has triggered, the condition's events no longer need
@@ -547,7 +585,7 @@ class Condition(Event):
             if isinstance(event, Condition):
                 event._remove_check_callbacks()
 
-    def _check(self, event):
+    def _check(self, event: Event) -> None:
         """Check if the condition was already met and schedule the *event* if
         so."""
         if self._value is not PENDING:
@@ -565,13 +603,13 @@ class Condition(Event):
             self.succeed()
 
     @staticmethod
-    def all_events(events, count):
+    def all_events(events: Tuple[Event, ...], count: int) -> bool:
         """An evaluation function that returns ``True`` if all *events* have
         been triggered."""
         return len(events) == count
 
     @staticmethod
-    def any_events(events, count):
+    def any_events(events: Tuple[Event, ...], count: int) -> bool:
         """An evaluation function that returns ``True`` if at least one of
         *events* has been triggered."""
         return count > 0 or len(events) == 0
@@ -583,8 +621,9 @@ class AllOf(Condition):
     any of *events* failed.
 
     """
-    def __init__(self, env, events):
-        super(AllOf, self).__init__(env, Condition.all_events, events)
+
+    def __init__(self, env: 'Environment', events: Iterable[Event]):
+        super().__init__(env, Condition.all_events, events)
 
 
 class AnyOf(Condition):
@@ -593,11 +632,12 @@ class AnyOf(Condition):
     any of *events* failed.
 
     """
-    def __init__(self, env, events):
-        super(AnyOf, self).__init__(env, Condition.any_events, events)
+
+    def __init__(self, env: 'Environment', events: Iterable[Event]):
+        super().__init__(env, Condition.any_events, events)
 
 
-def _describe_frame(frame):
+def _describe_frame(frame: FrameType) -> str:
     """Print filename, line number and function name of a stack frame."""
     filename, name = frame.f_code.co_filename, frame.f_code.co_name
     lineno = frame.f_lineno
@@ -607,5 +647,7 @@ def _describe_frame(frame):
             if no + 1 == lineno:
                 break
 
-    return '  File "%s", line %d, in %s\n    %s\n' % (filename, lineno, name,
-                                                      line.strip())
+    return (
+        f'  File "{filename}", line {lineno}, in {name}\n'
+        f'    {line.strip()}\n'
+    )
